/* 
 * Distort - BeOS screensaver based on the xscreensaver Distort
 * BeOS rewrite, copyright © 1999 by John Yanarella (yanarejm@muohio.edu)
 *
 * version 1.1 (March 29th, 2000)
 *    - updated to compile under R5
 *
 * version 1.0 (November 17th, 1999)
 *    - original release
 *
 * Changes from the original Distort xscreensaver:
 *  - moved from C -> C++, malloc -> new
 *  - rewritten for BeOS graphics, screensaver framework
 *	- added GUI configuration, custom lens type
 *  - added customizable twist for vortex, custom lens
 *	- fixes MANY drawing bugs in the original when speed > 1
 *  - eliminates the swamp_thing effect (WAY too resource hungry)
 *  - eliminates multiple lenses - bad bugs in edge detection & placement
 *    for large lens sizes; option may reappear in a future version
 *
 * Although the BeOS version involved a great deal of rewriting,
 * it's still covered by this license:
 * 
 * distort by Jonas Munsin (jmunsin@iki.fi) and Jamie Zawinski (jwz@jwz.org)
 *
 * xscreensaver, Copyright (c) 1992, 1993, 1994, 1996, 1997, 1998
 * Jamie Zawinski <jwz@jwz.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 *
 */

#include <Slider.h>
#include <TextView.h>
#include <Menu.h>
#include <MenuField.h>
#include <MenuItem.h>
#include <PopUpMenu.h>
#include <TabView.h>
#include <Screen.h>

#include <File.h>
#include <Node.h>
#include <NodeInfo.h>

#include <stdlib.h> // for srand(), random()
#include <math.h> // for sin(), cos(), atan2()

#include "Distort.h"

static const char* TITLE =
	"Distort for BeOS v1.1\n";
	
static const char* CREDITS = 
	"Bouncing Distortion Lens"
	"\n\nBeOS version by John Yanarella\n(yanarejm@muohio.edu)\n"
	"\nBased on the xscreensaver 'Distort'\n"
	"by Jonas Munsin (jmunsin@iki.fi),\n"
	"Randy Zack (randy@acucorp.com),\n"
	"and Jamie Zawinski (jwz@jwz.org).\n";

extern "C" _EXPORT BScreenSaver *instantiate_screen_saver(BMessage *message, image_id image)
{
	return new Distort(message, image);
}

Distort::Distort(BMessage *msg, image_id image)
 : BScreenSaver(msg, image)
{
	m_buffer_map = NULL;
	m_orig_map = NULL;
	m_from = NULL;
	m_icon = NULL;
	
	m_delay = 10000;

	m_image = image;
	m_preview = false;

	RestoreState(msg);

	if (m_type > 5)
		m_type = 0;
}

Distort::~Distort() 
{
	// see Distort::StopSaver() for deallocation
}

void Distort::StartConfig(BView *view)
{
	DistortConfigView* configView = new DistortConfigView(view->Bounds(), this);
	view->AddChild(configView);
}


status_t Distort::StartSaver(BView * view, bool preview)
{
	if (preview) {
		image_info info;
		BFile addon_file;
		if ((get_image_info(m_image, &info) == B_OK ) &&
			(addon_file.SetTo(info.name, O_RDONLY ) == B_OK))
		{
			BNodeInfo addon_node_info((BNode*)&addon_file);
			m_icon = new BBitmap(BRect(0,0,31,31), B_CMAP8);
			if (addon_node_info.GetTrackerIcon(m_icon, B_LARGE_ICON) == B_OK) {
				SetTickSize(1000000);
				m_preview = true;
				return B_OK;
			} else {
				delete m_icon;
				m_icon = NULL;
				return B_ERROR;
			}
		} else {
			delete m_icon;
			m_icon = NULL;
			return B_ERROR;
		}
	}

	// new seed every time we start
	time_t now;	
	now = time(NULL);
	srand(now);

	view->SetDrawingMode(B_OP_COPY);
	
	m_height = (int)view->Bounds().Height();
	m_width = (int)view->Bounds().Width();

	if (!(m_type == 3)) { // reflect doesn't currently use pre-calc'd lens
		allocate_lens();
		make_round_lens(m_radius,m_radius);
	}
	
	// initialize lens cooridinates
	m_xy_coo.x = (random() % (m_width-(2*m_radius+m_speed + 2)));
	m_xy_coo.y = (random() % (m_height-(2*m_radius+m_speed + 2)));	
	m_xy_coo.xmove = m_speed + 2*(-m_speed);
	m_xy_coo.ymove = m_speed + 2*(-m_speed);

	if (m_buffer_map)
		delete m_buffer_map;
	m_buffer_map = NULL;
	
	if (m_orig_map)
		delete m_orig_map;
	m_orig_map = NULL;

	BScreen screen(B_MAIN_SCREEN_ID);
	BRect screenframe = screen.Frame();
	m_colorspace = screen.ColorSpace();

	// check color space compatability
	if ((B_CMAP8 != m_colorspace) &&  (B_RGB15 != m_colorspace)
	&&  (B_RGB16 != m_colorspace) &&  (B_RGB32 != m_colorspace))
	{
		return B_ERROR;
	}
	
	// do a screen grab
	m_orig_map = new BBitmap(screenframe, m_colorspace);
	if (m_orig_map != NULL)
		screen.ReadBitmap(m_orig_map, true);
	else
		return B_ERROR;

	// create the buffer map
	BRect bufferframe(0, 0, 2*m_radius + 2*m_speed - 1, 2*m_radius + 2*m_speed - 1);
	m_buffer_map = new BBitmap(bufferframe, m_colorspace);
	if (m_buffer_map == NULL)
		return B_ERROR;
		
	// get pointers to the bitmap data
	m_orig8 = m_buf8 = NULL;
	m_orig16 = m_buf16 = NULL;
	m_orig32 = m_buf32 = NULL;
	switch (m_colorspace)
	{
		case B_RGB32:
			m_orig32 = (int32*) m_orig_map->Bits();
			m_buf32  = (int32*) m_buffer_map->Bits();
			break;

		case B_RGB15:
		case B_RGB16:
			m_orig16 = (int16*) m_orig_map->Bits();
			m_buf16  = (int16*) m_buffer_map->Bits();
			break;

		case B_CMAP8:
			m_orig8  = (int8*) m_orig_map->Bits();
			m_buf8   = (int8*) m_buffer_map->Bits();
			break;
		default:
			return B_ERROR;
	}

	SetTickSize(m_delay);

	return B_OK;
}

void Distort::StopSaver() {
	// delete icon used in preview
	if (m_icon != NULL)
		delete m_icon;	
	// free lens
	if (m_from != NULL)
		deallocate_lens();
	// free bitmaps
	if (m_buffer_map)
		delete m_buffer_map;
	if (m_orig_map)
		delete m_orig_map;
}

void Distort::Draw(BView *view, int32 frame)
{
	// preview draw
	if (m_preview) {
		if (frame == 0) {
			rgb_color color = { 102, 203, 255, 0 };
			view->SetHighColor(color);
			view->FillRect(view->Bounds());
			view->SetDrawingMode(B_OP_OVER);
			BPoint pt(view->Bounds().Width()/2 - 16, view->Bounds().Height()/2 - 16);
			view->DrawBitmap(m_icon, pt);
		}		
		return;
	}	
	
	// normal draw
	if (frame == 0) {
		view->DrawBitmap(m_orig_map);
	}
	
	move_lens();

	if (m_type == 3)
		reflect_draw(view);
	else
		plain_draw(view);	
	view->Sync();
}

status_t Distort::SaveState(BMessage *msg) const
{
	msg->AddInt32("type", m_type);
	msg->AddInt32("speed", m_speed);
	msg->AddInt32("radius", m_radius);
	msg->AddFloat("twist", m_twist);
	return B_OK;
}

void Distort::RestoreState(BMessage *msg) {
	// restore state from msg	
	int32 speed, radius, type;
	float twist;
	if (msg->FindInt32("speed", &speed) == B_OK)
		m_speed = speed;
	else
		m_speed = 5;
	if (msg->FindInt32("radius", &radius) == B_OK)
		m_radius = radius;
	else
		m_radius = 50;
	if (msg->FindInt32("type", &type) == B_OK)
		m_type = type;
	else
		m_type = 0;
	if (msg->FindFloat("twist", &twist) == B_OK)
		m_twist = twist;
	else
		m_twist = 3.5;
}

void Distort::allocate_lens()
{
	m_from = new int**[2*m_radius+2*m_speed];

	int i, j;		
	for (i = 0; i < 2*m_radius+2*m_speed; i++) {
		m_from[i] = new int*[2*m_radius+2*m_speed];
		for (j = 0; j < 2*m_radius+2*m_speed; j++) {
			m_from[i][j] = new int[2];
		}
	}
}

void Distort::deallocate_lens()
{
	int i, j;		
	for (i = 0; i < 2*m_radius+2*m_speed; i++) {		
		for (j = 0; j < 2*m_radius+2*m_speed; j++) {
			delete[] m_from[i][j];			
		}
		delete[] m_from[i];		
	}
	delete[] m_from;
}

// makes a lense with the Radius=loop and centered in the point (radius, radius)
void Distort::make_round_lens(int radius, int loop)
{
	int i, j, x, y;
	for (x = 0; x < 2*radius+2*m_speed; x++) {
		for(y = 0; y < 2*radius+2*m_speed; y++) {
			double r, d = 0.0;
			
			i = x - m_speed;
			j = y - m_speed;
						
			if (i >= 0 && j >= 0) {
				r = sqrt ((i-radius)*(i-radius)+(j-radius)*(j-radius));
				if (loop == 0)
					d = 0.0;
				else
					d = r/loop;
			} else
				r = loop;

			if (r < loop-1) {						
				if (m_type == 0) { // vortex-twist effect			

					// this one-line formula for getting a nice rotation angle 
					// is borrowed (with permission) from the whirl plugin for gimp,
		 			// Copyright (C) 1996 Federico Mena Quintero
					double angle = m_twist*(1-d)*(1-d);

					// avoid atan2
					if ((radius-j) == 0.0 && (radius-i) == 0.0) {
						m_from[x][y][0] = (int)(radius + cos(angle)*r) + m_speed;
						m_from[x][y][1] = (int)(radius + sin(angle)*r) + m_speed;
					} else {
      				  	m_from[x][y][0] = 
      				  		(int)(radius + cos(angle - atan2(radius-j, -(radius-i)))*r) + m_speed;
			        	m_from[x][y][1] = 
			        		(int)(radius + sin(angle - atan2(radius-j, -(radius-i)))*r) + m_speed;
					}
				} else { // default is to magnify (m_type 2)
					r = sin(d*M_PI_2);
				
	// raising r to different power here gives different amounts of
	// distortion, a negative value sucks everything into a black hole
	
					if (m_type == 1 && r != 0) // blackhole effect
						r = 1/r;

					if (m_type == 4)
						r = pow(r, m_twist);
					
					// bubble effect (and blackhole)
					m_from[x][y][0] = (int)((radius + (i-radius)*r) + m_speed);
					m_from[x][y][1] = (int)((radius + (j-radius)*r) + m_speed);					
				}
			} else { // not inside loop
				m_from[x][y][0] = x;
				m_from[x][y][1] = y;
			}
		}
	}
}

// generate an image of from[][][] and draw it on the screen
void Distort::plain_draw(BView* view)
{
	for(int i = 0; i < 2*m_radius+2*m_speed; i++) {
		for(int j = 0 ; j < 2*m_radius+2*m_speed ; j++) {
			if (m_xy_coo.x+m_from[i][j][0] >= 0 &&
				m_xy_coo.x+m_from[i][j][0] < m_width &&
				m_xy_coo.y+m_from[i][j][1] >= 0 &&
				m_xy_coo.y+m_from[i][j][1] < m_height)
			{
				switch (m_colorspace)
				{
					case B_RGB32:
						m_buf32[j*m_buffer_map->BytesPerRow()/4 + i] 
						 = 
						m_orig32[(m_xy_coo.y+m_from[i][j][1])
							     * m_orig_map->BytesPerRow()/4
								 + (m_xy_coo.x+m_from[i][j][0])];
						break;
				
					case B_RGB15:
					case B_RGB16:
						m_buf16[j*m_buffer_map->BytesPerRow()/2 + i]
						 =
						m_orig16[(m_xy_coo.y+m_from[i][j][1])
								* m_orig_map->BytesPerRow()/2
								+ (m_xy_coo.x+m_from[i][j][0])];
						
						break;
				
					case B_CMAP8:
						m_buf8[j*m_buffer_map->BytesPerRow() + i]
						 =
						m_orig8[(m_xy_coo.y+m_from[i][j][1])
								* m_orig_map->BytesPerRow()
								+ (m_xy_coo.x+m_from[i][j][0])];
						break;

					default:
						break;
				}
			} else {
					switch (m_colorspace)
				{
					case B_RGB32:
						m_buf32[j*m_buffer_map->BytesPerRow()/4 + i] = 0;	
						break;
				
					case B_RGB15:
					case B_RGB16:
						m_buf16[j*m_buffer_map->BytesPerRow()/2 + i] = 0;
						break;
				
					case B_CMAP8:
						m_buf8[j*m_buffer_map->BytesPerRow() + i] = 0;
						break;

					default:
						break;
				}
			}
		}
	}

	view->DrawBitmapAsync(m_buffer_map, BPoint(m_xy_coo.x, m_xy_coo.y));
}

// based on reflect algoritm by Randy Zack <randy@acucorp.com>
void Distort::reflect_draw(BView* view)
{
	int	cx, cy;
	int	ly, lysq, lx, ny, dist, rsq = m_radius * m_radius;

	cx = cy = m_radius + m_speed;

	for(int i = 0; i < 2*m_radius+2*m_speed; i++) {
		ly = i - cy;
		lysq = ly * ly;
		ny = m_xy_coo.y + i;
		for(int j = 0; j < 2*m_radius+2*m_speed; j++) {
			lx = j - cx;
			dist = lx * lx + lysq;
			if (dist > rsq || ly < -m_radius || ly > m_radius || lx < -m_radius || lx > m_radius)
			{					
				switch (m_colorspace)
				{
					case B_RGB32:
						m_buf32[i*m_buffer_map->BytesPerRow()/4 + j]
						 =
						m_orig32[(ny) * m_orig_map->BytesPerRow()/4 + (m_xy_coo.x + j)];
						break;
				
					case B_RGB15:
					case B_RGB16:
						m_buf16[i*m_buffer_map->BytesPerRow()/2 + j]
						 =
						m_orig16[(ny) * m_orig_map->BytesPerRow()/2 + (m_xy_coo.x + j)];
						break;
				
					case B_CMAP8:
						m_buf8[i*m_buffer_map->BytesPerRow() + j]
						 =
						m_orig8[(ny) * m_orig_map->BytesPerRow() + (m_xy_coo.x + j)];
						break;
					default:
						break;
				}
			} else if (dist == 0) {
				
				switch (m_colorspace)
				{
					case B_RGB32:
						m_buf32[i*m_buffer_map->BytesPerRow()/4 + j] = 0;
						break;
				
					case B_RGB15:
					case B_RGB16:
						m_buf16[i*m_buffer_map->BytesPerRow()/2 + j] = 0;
						break;
				
					case B_CMAP8:
						m_buf8[i*m_buffer_map->BytesPerRow() + j] = 0;
						break;
					default:
						break;
				}
				
			} else {
				
				int	x = m_xy_coo.x + cx + (lx * rsq / dist);
				int	y = m_xy_coo.y + cy + (ly * rsq / dist);
				if (x < 0 || x >= m_width || y < 0 || y >= m_height) {					
					switch (m_colorspace)
					{
						case B_RGB32:
							m_buf32[i*m_buffer_map->BytesPerRow()/4 + j] = 0;
							break;
					
						case B_RGB15:
						case B_RGB16:
							m_buf16[i*m_buffer_map->BytesPerRow()/2 + j] = 0;
							break;
					
						case B_CMAP8:
							m_buf8[i*m_buffer_map->BytesPerRow() + j] = 0;
							break;
						default:
							break;
					}
				}
				else {					
					switch (m_colorspace)
					{
						case B_RGB32:
							m_buf32[i*m_buffer_map->BytesPerRow()/4 + j]
							 =
							m_orig32[y*m_orig_map->BytesPerRow()/4 + x];
							break;
					
						case B_RGB15:
						case B_RGB16:
							m_buf16[i*m_buffer_map->BytesPerRow()/2 + j]
							 =
							m_orig16[y*m_orig_map->BytesPerRow()/2 + x];
							break;
					
						case B_CMAP8:
							m_buf8[i*m_buffer_map->BytesPerRow() + j]
							 =
							m_orig8[y*m_orig_map->BytesPerRow() + x];
							break;
						default:
							break;
					}
				}
			}
		}
	}

	view->DrawBitmapAsync(m_buffer_map, BPoint(m_xy_coo.x, m_xy_coo.y));
}

// move lens and handle bounces with walls
void Distort::move_lens()
{
	if (m_xy_coo.x + 2*m_radius + 2*m_speed >= m_width - m_speed)
		m_xy_coo.xmove = -abs(m_xy_coo.xmove);
	if (m_xy_coo.x <= m_speed) 
		m_xy_coo.xmove = abs(m_xy_coo.xmove);
	if (m_xy_coo.y + 2*m_radius + 2*m_speed >= m_height - m_speed)
		m_xy_coo.ymove = -abs(m_xy_coo.ymove);
	if (m_xy_coo.y <= m_speed)
		m_xy_coo.ymove = abs(m_xy_coo.ymove);

	m_xy_coo.x = m_xy_coo.x + m_xy_coo.xmove;
	m_xy_coo.y = m_xy_coo.y + m_xy_coo.ymove;
}

DistortConfigView::DistortConfigView(BRect frame, Distort *s)
	: BView(frame, "", B_FOLLOW_NONE, B_WILL_DRAW)
{
	m_screensaver = s;
	
	BTabView *tabView;
	tabView = new BTabView(frame, "tab_view");
	tabView->SetViewColor(216,216,216,0);
	
	frame.InsetBy(5,5); 
	frame.bottom -= tabView->TabHeight(); 

	// "credits" view
	BTextView* creditsView = new BTextView(frame, "credits", frame, B_FOLLOW_NONE, B_WILL_DRAW);
	
	creditsView->MakeEditable(false);
	creditsView->SetStylable(true);
	creditsView->MakeSelectable(false);	
	creditsView->SetAlignment(B_ALIGN_LEFT);
	creditsView->SetViewColor(216,216,216,0);
	
	BFont font(be_bold_font);
	font.SetSize(12.0);
	creditsView->SetFontAndColor(&font);
	creditsView->Insert(TITLE,strlen(TITLE));
	
	BFont font2(be_plain_font);
	font2.SetSize(10.0);
	creditsView->SetFontAndColor(&font2);	
	creditsView->Insert(CREDITS, strlen(CREDITS));

	BTab *tab;
	tab = new BTab(); 
	tabView->AddTab(creditsView, tab); 
	tab->SetLabel("Credits");
	
	BView *configView = new BView(frame, "", B_FOLLOW_NONE, B_WILL_DRAW);
	configView->SetViewColor(216,216,216,0); 

	tab = new BTab(); 
	tabView->AddTab(configView, tab); 
	tab->SetLabel("Config");

	BRect slideRect(configView->Bounds().InsetBySelf(5.0,0.0));
	rgb_color fill = { 0, 0, 255, 255 };

	// add "lens type" menu
	slideRect.top = 10;
	slideRect.bottom = 35;
	
	m_lens_menu = new BPopUpMenu("Lens type");
	BMenuItem *item[5];
	
	item[0] = new BMenuItem("Vortex", new BMessage(TYPE_CHANGED));
	m_lens_menu->AddItem(item[0]);
	item[1] = new BMenuItem("Blackhole", new BMessage(TYPE_CHANGED));
	m_lens_menu->AddItem(item[1]);
	item[2] = new BMenuItem("Magnify", new BMessage(TYPE_CHANGED));
	m_lens_menu->AddItem(item[2]);
	item[3] = new BMenuItem("Reflect", new BMessage(TYPE_CHANGED));
	m_lens_menu->AddItem(item[3]);
	item[4] = new BMenuItem("Custom", new BMessage(TYPE_CHANGED));
	m_lens_menu->AddItem(item[4]);
	if (m_screensaver->m_type > -1 && m_screensaver->m_type < 5)
		item[m_screensaver->m_type]->SetMarked(true);
	
	BMenuField *popup = new BMenuField(slideRect, "lensTypeField", "Lens type:", m_lens_menu);
	popup->SetDivider(60);
	configView->AddChild(popup);

	// "speed" slider
	slideRect.top = 40;
	slideRect.bottom = 90;
	m_speed_slider = new BSlider(slideRect, "speed", "Speed:",
	                           new BMessage(SPEED_CHANGED),
	                           1, 10, B_TRIANGLE_THUMB,
	                           B_FOLLOW_LEFT_RIGHT, B_WILL_DRAW|B_NAVIGABLE);
	m_speed_slider->SetLimitLabels("slow", "fast");
	m_speed_slider->SetHashMarks(B_HASH_MARKS_BOTTOM);
	m_speed_slider->SetHashMarkCount(10);
	m_speed_slider->SetKeyIncrementValue(1);
	m_speed_slider->UseFillColor(true, &fill);
	m_speed_slider->SetValue(m_screensaver->m_speed);
	configView->AddChild(m_speed_slider);	

	// "radius" slider
	slideRect.top = 100;
	slideRect.bottom = 150;
	m_radius_slider = new BSlider(slideRect, "radius", "Lens radius:",
	                           new BMessage(RADIUS_CHANGED),
	                           50, 150, B_TRIANGLE_THUMB,
	                           B_FOLLOW_LEFT_RIGHT, B_WILL_DRAW|B_NAVIGABLE);
	m_radius_slider->SetLimitLabels("small", "big");
	m_radius_slider->SetHashMarks(B_HASH_MARKS_BOTTOM);
	m_radius_slider->SetHashMarkCount(10);
	m_radius_slider->SetKeyIncrementValue(1);
	m_radius_slider->UseFillColor(true, &fill);
	m_radius_slider->SetValue(m_screensaver->m_radius);
	configView->AddChild(m_radius_slider);
	
	// "twist" slider
	slideRect.top = 160;
	slideRect.bottom = 210;
	m_twist_slider = new BSlider(slideRect, "twist", "Lens twist:",
	                           new BMessage(TWIST_CHANGED),
	                           -99, 99, B_TRIANGLE_THUMB,
	                           B_FOLLOW_LEFT_RIGHT, B_WILL_DRAW|B_NAVIGABLE);
	m_twist_slider->SetLimitLabels("-9.9", "9.9");
	m_twist_slider->SetHashMarks(B_HASH_MARKS_BOTTOM);
	m_twist_slider->SetHashMarkCount(11);
	m_twist_slider->SetKeyIncrementValue(1);
	m_twist_slider->UseFillColor(true, &fill);
	m_twist_slider->SetValue(m_screensaver->m_twist * 10);
	configView->AddChild(m_twist_slider);
	m_twist_slider->SetEnabled(m_screensaver->m_type == 0 || m_screensaver->m_type == 4);
	
	AddChild(tabView);
}

void DistortConfigView::AttachedToWindow()
{
	rgb_color view_color(Parent()->ViewColor());
	SetViewColor(view_color);
	m_speed_slider->SetTarget(this);
	m_radius_slider->SetTarget(this);
	m_twist_slider->SetTarget(this);
	m_lens_menu->SetTargetForItems(this);
}

void DistortConfigView::MessageReceived(BMessage *msg)
{
	BMenuItem *item;
	
	switch (msg->what) {
		case TYPE_CHANGED:
			m_screensaver->Lock();
			msg->FindPointer("source", (void **)&item);
			m_screensaver->m_type = m_lens_menu->IndexOf(item);
			m_twist_slider->SetEnabled(m_screensaver->m_type == 0 ||
									   m_screensaver->m_type == 4);	
			m_screensaver->Unlock();
			break;

		case SPEED_CHANGED:
			m_screensaver->Lock();
			m_screensaver->m_speed = m_speed_slider->Value();
			m_screensaver->Unlock();
			break;
			
		case RADIUS_CHANGED:
			m_screensaver->Lock();
			m_screensaver->m_radius = m_radius_slider->Value();
			m_screensaver->Unlock();
			break;

		case TWIST_CHANGED:
			m_screensaver->Lock();
			m_screensaver->m_twist = (float)m_twist_slider->Value()/10.0;
			m_screensaver->Unlock();
			break;
			
		default:
			BView::MessageReceived(msg);
			break;
	}
}